#ifndef __TEXTOUTPUT_CPP__
#define __TEXTOUTPUT_CPP__

#include <map>
#include <string>
#include <utility>

const int MAX_BUF_SIZE = 1000;

#define TEXT_MODE_NONE 0
#define TEXT_MODE_FLAT 1
#define TEXT_MODE_POSITIONED 2

#ifndef TEXT_MODE
#  ifndef DEMO_MODE
#    define TEXT_MODE TEXT_MODE_POSITIONED
#  else
#    define TEXT_MODE TEXT_MODE_NONE
#  endif
#endif

/** Print formatted text at a certain (x,y) position on the console.
 * To write in the standard mode (i.e. text is not positioned, but every time a new line is inserted),
 * define TEXT_MODE to TEXT_MODE_FLAT. To avoid writing, define TEXT_MODE to TEXT_MODE_NONE.
 */
class Textout
{
public:
  /** 
   * Prints the string s at position (line,column) in the console.
   */
  static void print(int line, int column, const std::string& s, bool nolinebreak = false);

  /**
   * Prints the formatted string provided with format and a variable number of arguments to the console @see print().
   * The function is similar to printf, and so are the formatting parameters in format. 
   */
  static void printvar(int line, int column, const std::string& format, ...);

  /**
   * Inserts a new page in the console. 
   * If the console contains 80 lines, then 80 new lines are inserted and writing continues at position 80 + line.
   */
  static void pagebreak();

  /**
   * Prints a visual line (--------) at position line that spans over the whole width of the console.
   */
  static void printbreak(int line);

  /**
   * Prints a busy cursor at position (line, column).
   *
   * Each time printbusy(line,column) is called for a certain position, the character at this position changes from
   * / -> - -> \ -> | -> ... This provides a 'console busy clock'. 
   */
  static void printbusy(int line, int column);

  /**
   * Clears the whole console window.
   */
  static void clear();

  /**
   * Creates and initializes the Textout-class and its associated console window.
   *
   * This function must be called exactly once on initializing the application.
   */
  static void init();

private:
  static char    text[MAX_BUF_SIZE];
  static int     curLine;
  static int     bufferwidth;
  static std::map<std::pair<int, int>,int> busyStates;
};

#endif
